"""Qualia-Plugin-SNN preprocessing package contains preprocessing modules adapted for or dedicated to Spiking Neural Networks."""

from .Split2TimeSteps import Split2TimeSteps

__all__ = ['Split2TimeSteps']
